#
#  ezm/tk GUI definition by tcl/tk
#
set ezm_version "0.07 <1997/11/18>"

# initialization
set username {}
set password {}
set n_mail 0
set mailfile {}
set popsock {}
set current_mail -1
set reading_spool 0
set pophost {}
set pophosts {}
set genericfrom 0
set nimage 0
set image_list {}
set image_file_list {}

# debugging switches
set debug 0
set tcl_only 0
#set prog_base c:/soft/ezmdev/

######## DEBUG
if {$tcl_only} {
    proc jis2sj {s} {return $s}
    proc sj2jis {s} {return $s}
    proc f_unlink {s} {exec rm $s}
    proc f_rename {s1 s2} {exec mv $s1 $s2}
    proc f_mkdir {d} {}
    proc str_equal {x y} {return [expr [string compare $x $y] == 0]}
    proc str_equal_nocase {x y} {return [expr [string compare -nocase $x $y] == 0]}
    proc head_match {h x} { return [regexp "^$h" $x]}
    proc head_match_nocase {h x} { return [regexp -nocase "^$h" $x]}
    proc encode_mime {x} {return $x}
    proc decode_mime {x} {return $x}
    proc mailaddr_equal {x y} {return [expr [string compare -nocase $x $y] == 0]}
}
##################################################################

# load package
if [str_equal $tcl_platform(platform) windows] {
    # Windows 95/NT
    source "${prog_base}wincfg.tcl"

    lappend auto_path $prog_base
    proc jis2internal {s} {return [jis2sj $s]}
    proc internal2jis {s} {return [sj2jis $s]}
} else {
    # UNIX + X window system
    source "${prog_base}unixcfg.tcl"
    set internal SJIS
    proc jis2internal {s} { 
        return [kanji conversion JIS SJIS $s] 
    }
    proc internal2jis {s} { 
        global internal
        return [kanji conversion $internal JIS $s] 
    }
    kanji internalCode SJIS
}

# read POP hosts file
set pophostsfile "${prog_base}pophosts.ini"
if [catch {set f [open $pophostsfile r]}] {
    tk_messageBox -type ok -icon error -message {POPzXg̐ݒt@C܂}
} else {
    while {! [eof $f]} {
        gets $f line
        if {[string length $f] == 0} { continue }
        lappend pophosts $line
    }
    close $f
}

proc alias {name body} {
    global aliases
    set aliases($body) $name
}
proc mlalias {name body} {
    global mlaliases
    set mlaliases($body) $name
}

# read rc file

if {[file exists $rcfile]} { source $rcfile }

# create mail folder if needed
if {[str_equal $tcl_platform(platform) windows] && [regexp {^[A-Za-z]:$} $mailbox]} {
    # Windows drive letter is not a directory
} elseif {! [file isdirectory $mailbox]} {
    f_mkdir $mailbox
}

#
# command procs ----------------------------------------------------------
#

# read mail file
proc cmd_read_file {} {
    global reading_spool mailbox n_mail
    if {$reading_spool} {
        update_spool
        set reading_spool 0
    }
    clear_mail
    read_mail [tk_getOpenFile -initialdir $mailbox -filetypes {
                  {"Text file" {".txt"}} {"All file" {"*"}}}]
    focus -force .tl
    if {$n_mail > 0} {
        select_article 0
    }
}

# save current mail to selected file

proc cmd_save_old {} {
    global current_mail
    if [m_save {}] {
        article_status $current_mail saved
    }
}

# quit
proc cmd_quit {} {
    update_spool
    exit 0
}

# move to next/previous mail
proc cmd_next_mail {} {
    global current_mail n_mail
    if {$current_mail >= 0 && $current_mail < $n_mail} {
	incr current_mail
	set y [expr $current_mail-1]
	tl_scroll see $y
        select_article $y
    }
}
proc cmd_previous_mail {} {
    global current_mail
    if {$current_mail > 0} {
	incr current_mail -1
	set y [expr $current_mail-1]
	tl_scroll see $y
        select_article $y
    }
}


# copy region to the clipboard
proc cmd_copy_clipboard {} {
    clipboard clear -displayof .tx.text 
    clipboard append -displayof .tx.text [.tx.text get sel.first sel.last]
}

# receive new mails
proc receive_mail {} {
    global pophost spoolfile popsock reading_spool n_mail
    if [catch {set sp [open $spoolfile a]}] {
        cantopen $spoolfile
        return
    }
    close $sp
    if {[string length $popsock] == 0} {
	pop_open
	if {[string length $popsock] == 0} { return }
    }
    if {! $reading_spool} {
	clear_mail
    } else {
        update_spool
	clear_mail
    }
    pop_read_spool
    set reading_spool 1
    read_mail $spoolfile
    if {$n_mail > 0} {
        select_article 0
    }
}

# query username/password again and receive mail
proc cmd_connect {} {
    global password
    set password {}
    receive_mail
}

# show version info
proc show_version {} {
    global ezm_version
    tk_messageBox -icon info -type ok -message "EZM version $ezm_version by Akinori Ito (aito@ei5sun.yz.yamagata-u.ac.jp)"
}

# save current mail
proc cmd_save_mail {} {
    global current_mail
    set f [.tb.e get]
    if {[string length f] == 0} {
	tk_messageBox -icon error -type ok -message {ۑ悪w肳Ă܂}
    } elseif {[m_save $f]} {
	article_status $current_mail saved
    }
}

# mark current mail as deleted
proc cmd_delete_mail {} {
    global current_mail mail_status
    set f [.tb.e get]
    if {$current_mail == -1} {
	tk_messageBox -icon error -type ok -message {܂[Ił}
    } else {
        if [str_equal $mail_status([expr $current_mail-1]) deleted] {
	    article_status $current_mail processed
        } else {
	    article_status $current_mail deleted
        }
    }
}

# SlistboxxɃXN[
proc tl_scroll args {
    foreach widget {num from to subject} {
	eval [join [list ".tl.$widget " $args]]
    }
}
# Slistboẍʒu킹
proc tl_sync {} {
    set pos [.tl.v get]
    tl_scroll yview moveto [lindex $pos 0]
}

proc tl_select {y} {
    foreach widget {from to subject} {
        .tl.$widget selection clear 0 end
	.tl.$widget selection set $y $y
    }
}
proc select_article {y} {
    global current_mail
    set current_mail -1
    tl_select $y
    update
    regsub -all {[^0-9]*} [.tl.num get $y] {} current_mail
    if {[string length $current_mail] > 0 && $current_mail >= 0} {
	show_mail $current_mail
    } else {
	set current_mail -1
    }
}


#
# Create GUI widgets -----------------------------------------------------
#

. configure -bg $bgcolor 
bind . <Destroy> { 
    update_spool
}        

#### MENU BAR
frame .mb -bg $bgcolor -relief raise -borderwidth 1
menubutton .mb.file -bg $bgcolor -text "t@C" -font $font -menu .mb.file.m
menubutton .mb.edit -bg $bgcolor -text "ҏW"     -font $font -menu .mb.edit.m
menubutton .mb.mail -bg $bgcolor -text "莆"     -font $font -menu .mb.mail.m
menubutton .mb.conf -bg $bgcolor -text "ݒ"     -font $font -menu .mb.conf.m
menubutton .mb.help -bg $bgcolor -text "wv"   -font $font -menu .mb.help.m
image create photo ezm_mark -file "${prog_base}ezm.gif"
label .mb.mark -image ezm_mark -borderwidth 0
#
pack .mb -side top -fill both -anchor w
pack .mb.file .mb.edit .mb.mail .mb.help -side left
pack .mb.mark .mb.conf -side right

#### MENUS
menu .mb.file.m -tearoff 0 -bg $bgcolor 
.mb.file.m add command -label "t@Cǂ" -font $font -command cmd_read_file

.mb.file.m add command -label "t@Cw肵ĕۑ..." -font $font \
	-command cmd_save_old
.mb.file.m add command -label "" -font $font -command { print_mail }
.mb.file.m add separator
.mb.file.m add command -label "I" -font $font -command cmd_quit
#
menu .mb.edit.m -tearoff 0 -bg $bgcolor 
#.mb.edit.m add command -font $font -label "؂"
.mb.edit.m add command -font $font -label {Rs[} -command cmd_copy_clipboard
#.mb.edit.m add command -label {\t} -font $font -command { \
	#    .tx.text insert insert [selection get -selection CLIPBOARD] }
#
menu .mb.mail.m -tearoff 0 -bg $bgcolor 
.mb.mail.m add command -label "M" -font $font -command receive_mail
.mb.mail.m add separator
.mb.mail.m add command -label "ԐM" -font $font -command {send_mail reply}
.mb.mail.m add command -label "VM" -font $font -command {send_mail new}
.mb.mail.m add command -label "]" -font $font -command {send_mail forward}
.mb.mail.m add separator
.mb.mail.m add command -label "ĐڑĎM" -font $font -command cmd_connect
#
menu .mb.help.m -tearoff 0 -bg $bgcolor 
.mb.help.m add command -label "ǂ" -font $font -command {
    popup_panel .help
    grab release .help
}
.mb.help.m add command -label {o[W} -font $font -command show_version
#
menu .mb.conf.m -tearoff 0 -bg $bgcolor
.mb.conf.m add command -label {ʖ̐ݒ} -font $font -command edit_alias
.mb.conf.m add checkbutton -label {Generic From `} -variable genericfrom \
                           -font $font -command { update_ini }
.mb.conf.m add command -label {POPzXg̕ҏW} -font $font -command edit_pophosts
#

#### SAVE FILE PANE
frame .tb -bg $bgcolor
label .tb.l -bg $bgcolor -font $font -text "ۑ:"
entry .tb.e -bg $bgcolor -width 60 -font $font
button .tb.ok -bg $bgcolor -font $font -text {ۑ} -command cmd_save_mail
button .tb.del -bg $bgcolor -font $font -text {폜} -command cmd_delete_mail
pack .tb -side top -anchor w
grid .tb.l -column 0 -row 0
grid .tb.e -column 1 -row 0 -sticky ew
grid .tb.ok -column 2 -row 0
grid .tb.del -column 3 -row 0
grid columnconfigure .tb 1 -weight 100 

bind .tb.e <Return> { 
    cmd_save_mail
    focus .tl
}

#### SUMMARY PANE
frame .tl -bg $bgcolor
listbox .tl.num -width 4 -height 6 -font $font -setgrid true -borderwidth 0\
	-bg $bgcolor -relief flat -yscrollcommand ".tl.v set"
listbox .tl.from -width 19 -height 6 -font $font -setgrid true \
	-bg $bgcolor -yscrollcommand ".tl.v set"
listbox .tl.to -width 19 -height 6 -font $font -setgrid true \
	-bg $bgcolor -yscrollcommand ".tl.v set"
listbox .tl.subject -width 36 -height 6 -font $font -setgrid true \
	-bg $bgcolor -yscrollcommand ".tl.v set"
scrollbar .tl.v -command "tl_scroll yview" -bg $bgcolor 
label .tl.l1 -text FROM -bg $bgcolor 
label .tl.l2 -text TO -bg $bgcolor 
label .tl.l3 -text SUBJECT -bg $bgcolor 
pack .tl -side top -fill both
grid .tl.l1 -column 1 -row 0
grid .tl.l2 -column 2 -row 0
grid .tl.l3 -column 3 -row 0
grid .tl.num -column 0 -row 1 -sticky ns
grid .tl.from -column 1 -row 1 -sticky ns
grid .tl.to -column 2 -row 1 -sticky ns
grid .tl.subject -column 3 -row 1 -sticky nsew
grid .tl.v -column 4 -row 1 -sticky ns
grid columnconfigure .tl 3 -weight 100

foreach widget {num from to subject} {
    bind .tl.$widget <Button-1> {
        select_article [.tl.from nearest %y]
        focus .tl
    }
} 

#### BROWSER PANE
frame .tx -bg $bgcolor
text .tx.text -width 80 -height 13 -font $font -setgrid true \
        -bg $bgcolor -state disabled\
	-yscrollcommand ".tx.v set" -xscrollcommand ".tx.h set"
scrollbar .tx.v -bg $bgcolor -orient vertical   -command ".tx.text yview"
scrollbar .tx.h -bg $bgcolor -orient horizontal -command ".tx.text xview"
pack .tx -side top -fill both -expand 1
grid .tx.text -column 0 -row 0 -sticky nsew
grid .tx.v -column 1 -row 0 -sticky ns
grid .tx.h -column 0 -row 1 -sticky ew
grid rowconfigure .tx 0 -weight 100
grid rowconfigure .tx 1 -weight 1
grid columnconfigure .tx 0 -weight 100

# resize separetors in the text window
bind .tx.text <Configure> {
    set sep_width [expr [winfo width .tx.text]-10]
    foreach w [.tx.text window names] {
        if [head_match $w ".sep"] {
            $w configure -width $sep_width
        }
    }
}

#
# Key bindings --------------------------------------------------------------
#
bind .tl <Key-space>         {.tx.text yview scroll +1 page}
#bind .tl <Key-Page_Up>       {.tx.text yview scroll +1 page}
bind .tl <Key-b>             {.tx.text yview scroll -1 page}
bind .tl <Key-BackSpace>     {.tx.text yview scroll -1 page}
#bind .tl <Key-Page_Down>     {.tx.text yview scroll -1 page}
bind .tl <Key-n>             cmd_next_mail
bind .tl <Key-p>             cmd_previous_mail
bind .tl <Key-Up>            {.tx.text yview scroll -1 unit}
bind .tl <Key-Down>          {.tx.text yview scroll +1 unit}
bind .tl <Key-s>             cmd_save_mail
bind .tl <Key-d>             cmd_delete_mail
bind .tl <Key-r>             {send_mail reply}
bind .tl <Key-a>             {send_mail new}
bind .tl <Key-f>             {send_mail forward}
bind .tl <Key-R>             receive_mail
bind .tl <Key-q>             cmd_quit
bind .tl <Key-o>             cmd_read_file
bind .tl <Key-S>             cmd_save_old

# initial focus
focus .tl

# debug window
if {$debug} {
    toplevel .debug
    text .debug.win -font $font -height 50
    pack .debug.win
}

#--------------------------------------------------------------------------

#
# Username/Password query panel --------------------------------------------
#
set uname_ok 0
set create_method(.uname) create_uname
proc create_uname {} {
    global uname_ok font bgcolor pophost pophosts
    toplevel .uname
    wm withdraw .uname
    wm transient .uname .
    label .uname.lh -text {POPzXg:} -font $font
    label .uname.l0 -text {[U:} -font $font
    label .uname.l1 -text {pX[h:} -font $font
    if {[string length $pophost] == 0} {
        set pophost [lindex $pophosts 0]
    } else {
        set pophost_not_found 1
        foreach h $pophosts {
            if [str_equal $h $pophost] {
                set pophost_not_found 0
                break
            }
        }
        if {$pophost_not_found} {
            lappend pophosts $pophost
        }
    }
    eval [join [concat tk_optionMenu .uname.ph pophost $pophosts]]
    entry .uname.p0 -width 30 -font $font
    bind .uname.p0 <Return> { focus -force .uname.p1 }
    bind .uname.p0 <Tab> { focus -force .uname.p1 }
    entry .uname.p1 -width 30 -font $font -show *
    bind .uname.p1 <Return> { event generate .uname <<uname_ok>> }
    bind .uname.p1 <Tab> { event generate .uname <<uname_ok>> }
    frame .uname.b
    button .uname.b.ok -text "OK" -font $font  \
	    -command { event generate .uname <<uname_ok>> }
    button .uname.b.cancel -text {LZ} -font $font \
	    -command { event generate .uname <<uname-cancel>> }
    bind .uname <<uname_ok>> {
	set uname_ok 1
	wm withdraw .uname
    }
    bind .uname <<uname-cancel>> {
	set uname_ok 0
	wm withdraw .uname
    }

    grid .uname.lh -column 0 -row 0 -sticky w
    grid .uname.l0 -column 0 -row 1 -sticky w
    grid .uname.l1 -column 0 -row 2 -sticky w
    grid .uname.ph -column 1 -row 0 -sticky ew
    grid .uname.p0 -column 1 -row 1 -sticky ew
    grid .uname.p1 -column 1 -row 2 -sticky ew
    grid .uname.b -column 0 -row 3 -columnspan 2 -sticky nw
    pack .uname.b.ok -side left
    pack .uname.b.cancel -side left
}

# Query username and password
proc query_userpass {} {
    global uname_ok username password pophost
    popup_panel .uname
    focus -force .uname.p0
    .uname.p0 delete 0 end
    .uname.p1 delete 0 end
    if {[string length $username] > 0} {
        .uname.p0 insert 0 $username
    }
    set old_pophost $pophost
    tkwait variable uname_ok
    popdown_panel .uname
    if {$uname_ok} {
	set username [.uname.p0 get]
	set password [.uname.p1 get]
        if {![str_equal $pophost $old_pophost]} { update_ini }
    } else {
	set username {}
	set password {}
    }
}
#--------------------------------------------------------------------------

#
# To/Cc/Subject query panel --------------------------------------------
#
set sendquery_ok 0
set create_method(.sendquery) create_sendquery
proc create_sendquery {} {
    global font sendquery_ok
    toplevel .sendquery
    wm withdraw .sendquery
    wm transient .sendquery .
    label .sendquery.l0 -text {To:} -font $font
    label .sendquery.l1 -text {Cc:} -font $font
    label .sendquery.l2 -text {Subject:} -font $font
    entry .sendquery.p0 -width 30 -font $font
    bind .sendquery.p0 <Return> { focus -force .sendquery.p1 }
    bind .sendquery.p0 <Tab> { focus -force .sendquery.p1 }
    entry .sendquery.p1 -width 30 -font $font
    bind .sendquery.p1 <Return> { focus -force .sendquery.p2 }
    bind .sendquery.p1 <Tab> { focus -force .sendquery.p2 }
    entry .sendquery.p2 -width 30 -font $font
    bind .sendquery.p2 <Return> { event generate .sendquery <<sendquery_ok>> }
    bind .sendquery.p2 <Tab> { event generate .sendquery <<sendquery_ok>> }

    frame .sendquery.b
    button .sendquery.b.ok -text "OK" -font $font  \
	    -command { event generate .sendquery <<sendquery_ok>> }
    button .sendquery.b.cancel -text {LZ} -font $font \
	    -command { event generate .sendquery <<sendquery-cancel>> }
    bind .sendquery <<sendquery_ok>> {
	global sendquery_ok
	set sendquery_ok 1
	wm withdraw .sendquery
    }
    bind .sendquery <<sendquery-cancel>> {
	global sendquery_ok
	set sendquery_ok 0
	wm withdraw .sendquery
    }

    grid .sendquery.l0 -column 0 -row 0 -sticky w
    grid .sendquery.l1 -column 0 -row 1 -sticky w
    grid .sendquery.l2 -column 0 -row 2 -sticky w
    grid .sendquery.p0 -column 1 -row 0
    grid .sendquery.p1 -column 1 -row 1
    grid .sendquery.p2 -column 1 -row 2
    grid .sendquery.b -column 0 -row 3 -columnspan 2 -sticky nw
    pack .sendquery.b.ok -side left
    pack .sendquery.b.cancel -side left
}
#--------------------------------------------------------------------------

#
# Resent-To: query panel --------------------------------------------
#
set fwdquery_ok 0
set create_method(.fwdquery) create_fwdquery
proc create_fwdquery {} {
    global font fwdquery_ok
    toplevel .fwdquery
    wm withdraw .fwdquery
    wm transient .fwdquery .
    label .fwdquery.l0 -text {]:} -font $font
    entry .fwdquery.p0 -width 30 -font $font
    bind .fwdquery.p0 <Return> { event generate .fwdquery <<fwdquery_ok>> }
    bind .fwdquery.p0 <Tab> { event generate .fwdquery <<fwdquery_ok>> }

    frame .fwdquery.b
    button .fwdquery.b.ok -text "OK" -font $font  \
	    -command { event generate .fwdquery <<fwdquery_ok>> }
    button .fwdquery.b.cancel -text {LZ} -font $font \
	    -command { event generate .fwdquery <<fwdquery-cancel>> }
    bind .fwdquery <<fwdquery_ok>> {
	global fwdquery_ok
	set fwdquery_ok 1
	wm withdraw .fwdquery
    }
    bind .fwdquery <<fwdquery-cancel>> {
	global fwdquery_ok
	set fwdquery_ok 0
	wm withdraw .fwdquery
    }

    grid .fwdquery.l0 -column 0 -row 0 -sticky w
    grid .fwdquery.p0 -column 1 -row 0
    grid .fwdquery.b -column 0 -row 1 -columnspan 2 -sticky nw
    pack .fwdquery.b.ok -side left
    pack .fwdquery.b.cancel -side left
}
#--------------------------------------------------------------------------
#
# progress bar ------------------------------------------------------------
#
set create_method(.progress) create_progress
set prog_ratio 1.0
set prog_id {}
proc create_progress {} {
    global prog_ratio prog_id font bgcolor
    toplevel .progress -bg $bgcolor
    wm withdraw .progress
    wm transient .progress .
    label .progress.l -font $font -bg $bgcolor
    canvas .progress.c -width 400 -height 20
    pack .progress.l -side top -fill both
    pack .progress.c -side top
    set prog_id [.progress.c create rectangle 0 0 400 20 -fill #ffff00 -outline #ffff00]
}

proc progb_ratio {r} {
    global prog_id prog_ratio
    if {$r == 0} { 
	set scale_factor 0.001
	set prog_ratio [expr $prog_ratio * 0.001] 
    } else {
	set scale_factor [expr $r / $prog_ratio]
	set prog_ratio $r
    }
    .progress.c scale $prog_id 0 0 $scale_factor 1
    update
}

proc progb_message {m} {
    .progress.l configure -text $m
    update
}

# -------------------------------------------------------------------------

# Popup for message -------------------------------------------------------
set create_method(.mesg) create_mesg
proc create_mesg {} {
    global font bgcolor
    toplevel .mesg
    wm withdraw .mesg
    wm transient .mesg .
    label .mesg.l -font $font -bg $bgcolor
    pack .mesg.l
}
proc show_mesg {message} {
    .mesg.l configure -text $message
    update
}

# -------------------------------------------------------------------------

# Help window       -------------------------------------------------------

set help_nsection 0
set help_currentsection 0
set create_method(.help) create_help
proc create_help {} {
    global font bgcolor prog_base dir_separator helpindex_pos help_nsection help_currentsection
    set helpfile [join [list $prog_base  "readme.txt"] {}]
    if {! [file exists $helpfile]} {
        cantopen $helpfile
        return
    }
    set h [open $helpfile r]
    toplevel .help
    wm title .help {EZM help viewer}
    frame .help.index -bg $bgcolor
    pack .help.index -side top -fill both
    button .help.index.dismiss -font $font -text {} \
        -bg $bgcolor -command {
        popdown_panel .help
    }
    listbox .help.index.x -width 40 -height 4 -font $font -bg $bgcolor \
        -yscrollcommand {.help.index.v set}
    scrollbar .help.index.v -command ".help.index.x yview" -bg $bgcolor
    image create photo ezm_bmark -file "${prog_base}ezmb.gif"
    button .help.index.mark -image ezm_bmark -borderwidth 0 -command show_version
    grid .help.index.dismiss -row 0 -column 0 -sticky nsew
    grid .help.index.x -row 0 -column 1 -sticky ew
    grid .help.index.v -row 0 -column 2 -sticky ns
    grid .help.index.mark -row 0 -column 3
    grid columnconfigure .help.index 1 -weight 100
    set i 0
    while {! [eof $h]} {
        set pos [tell $h]
        gets $h line
        if [regexp {^\*[0-9]} $line] {
             set helpindex_pos($i) $pos
             .help.index.x insert $i "$line"
             incr i
        }
    }
    close $h
    set help_nsection $i
    bind .help.index.x <Button-1> {
        set help_currentsection [.help.index.x nearest %y]
        show_help $help_currentsection
        focus -force .help.index
    }
    bind .help.index <Key-Down> {
        if {$help_currentsection < [expr $help_nsection-1]} {
            incr help_currentsection
            .help.index.x see $help_currentsection
            .help.index.x selection clear 0 end
            .help.index.x selection set $help_currentsection $help_currentsection
            show_help $help_currentsection
        }
    }
    bind .help.index <Key-Up> {
        if {$help_currentsection > 0} {
            incr help_currentsection -1
            .help.index.x see $help_currentsection
            .help.index.x selection clear 0 end
            .help.index.x selection set $help_currentsection $help_currentsection
            show_help $help_currentsection
        }
    }
    bind .help.index <Key-space> {
        .help.b.x yview scroll +1 page
    }
    frame .help.b 
    pack .help.b -side top -expand 1 -fill both
    text .help.b.x -width 80 -height 20 -font $font -bg $bgcolor \
        -yscrollcommand {.help.b.v set}
    scrollbar .help.b.v -command ".help.b.x yview" -bg $bgcolor
    pack .help.b.x -side left -fill both -expand 1
    pack .help.b.v -side left -fill y
}

proc show_help {pos} {
    global prog_base helpindex_pos
    set helpfile [join [list $prog_base "readme.txt"] {}]
    if {! [file exists $helpfile]} {
        cantopen $helpfile
        return
    }
    set h [open $helpfile r]
    .help.b.x configure -state normal
    .help.b.x delete 1.0 end
    seek $h $helpindex_pos($pos)
    gets $h line
    .help.b.x insert end "$line\n"
    while {! [eof $h]} {
        gets $h line
        if [head_match $line "*"] { break }
        .help.b.x insert end "$line\n"
    }
    close $h
    .help.b.x configure -state disabled
}


# -------------------------------------------------------------------------

# Printer selector   -------------------------------------------------------

if [str_equal $tcl_platform(platform) windows] {
proc getprinters {winini} {
    global printers
    set printers(list) {}
    set printers(default) {}
    if [catch {set f [open $winini r]}] {
        return
    }
    while {! [eof $f]} {
        gets $f line
        if {[string compare $line {[PrinterPorts]}] == 0} {
             while {! [eof $f]} {
                gets $f line
                if {[string length $line] == 0 || [regexp {^\[} $line]} {
                    break
                }
                if [regexp {(.+)=.*} $line dummy printername] {
                     lappend printers(list) $printername
                }
            }
        }
        if {[string compare $line {[windows]}] == 0} {
             while {! [eof $f]} {
                gets $f line
                if {[string length $line] == 0 || [regexp {^\[} $line]} {
                    break
                }
                if [regexp {device=([^,]+),} $line dummy printername] {
                     set printers(default) $printername
                }
            }
        }
    }
    close $f
}

set choosed_printer {}
set create_method(.prselect) create_prselect
set prselect_ok {}
proc create_prselect {} {
    global bgcolor font choosed_printer
    toplevel .prselect
    wm withdraw .prselect
    wm transient .prselect
    label .prselect.l -bg $bgcolor -font $font -text {v^Ił}
    pack .prselect.l -side top -fill x 
    frame .prselect.f1
    listbox .prselect.f1.l -bg $bgcolor -font $font -width 30 -height 5 -yscrollcommand {.prselect.f1.v set}
    scrollbar .prselect.f1.v -bg $bgcolor -orient vertical -command {.prselect.f1.l yview}
    pack .prselect.f1 -side top
    pack .prselect.f1.l -side left -fill y
    pack .prselect.f1.v -side left -fill y
    frame .prselect.f2
    bind .prselect.f1.l <Button-1> {
        set choosed_printer [.prselect.f1.l nearest %y]
    }

    button .prselect.f2.ok -text "OK" -font $font -bg $bgcolor \
	    -command {
        set prselect_ok 1
    }
    button .prselect.f2.cancel -text {LZ} -font $font -bg $bgcolor \
	    -command {
        set prselect_ok 0
    }
    pack .prselect.f2 -side top
    pack .prselect.f2.ok .prselect.f2.cancel -side left 
}

proc select_printer {} {
    global prselect_ok choosed_printer win_ini printers
    getprinters $win_ini
    if {[llength $printers(list)] == 0} {
        tk_messageBox -icon error -type ok -message {v^܂}
        return {}
    }
    popup_panel .prselect
    set i 0
    .prselect.f1.l delete 0 end
    foreach p $printers(list) {
        .prselect.f1.l insert $i $p
        if {[string compare $p $printers(default)] == 0} {
            .prselect.f1.l selection set $i
            set choosed_printer $i
        }
        incr i
    }
    tkwait variable prselect_ok
    popdown_panel .prselect
    if {$prselect_ok && $choosed_printer > 0} {
        return [lindex $printers(list) $choosed_printer]
    }
    return {}
}
}

# POP host editor   -------------------------------------------------------
set create_method(.popedit) create_popedit
set popedit_ok 0
proc create_popedit {} {
    global font bgcolor popedit_ok
    toplevel .popedit -bg $bgcolor
    label .popedit.lbl -bg $bgcolor -font $font -text {POPzXg̕ҏW}
    text .popedit.e -width 40 -height 5 -yscrollcommand ".popedit.v set" \
                    -bg $bgcolor
    scrollbar .popedit.v -orient vertical -command ".popedit.e yview" \
              -bg $bgcolor
    frame .popedit.b -bg $bgcolor
    button .popedit.b.ok -command {set popedit_ok 1} -text {OK} \
                         -font $font -bg $bgcolor
    button .popedit.b.cancel -command {set popedit_ok 0} -text {LZ} \
                         -font $font -bg $bgcolor
    grid .popedit.lbl -column 0 -row 0 -columnspan 2 -sticky ew
    grid .popedit.e -column 0 -row 1 -sticky nsew
    grid .popedit.v -column 1 -row 1 -sticky ns
    grid .popedit.b -column 0 -row 2 -columnspan 2 -sticky ew
    pack .popedit.b.ok -side left
    pack .popedit.b.cancel -side left
}

proc edit_pophosts {} {
    global popedit_ok pophosts pophostsfile
    popup_panel .popedit
    .popedit.e delete 1.0 end
    foreach h $pophosts {
        .popedit.e insert end "$h\n"
    }
    tkwait variable popedit_ok
    popdown_panel .popedit
    if {$popedit_ok} {
        set lastline [.popedit.e index end]
        set i 1
        set pophosts {}
        while {$i < $lastline} {
            lappend pophosts [.popedit.e get $i.0 $i.end]
            incr i
        }
        if [catch {set f [open $pophostsfile w]}] {
            return
        }
        foreach h $pophosts {
            puts $f $h
        }
        close $f
    }
}

# alias editor   -----------------------------------------------------------
set create_method(.alias) create_aliasedit
set alias_ok 0
set selected_alias_entry {}
proc create_aliasedit {} {
    global font bgcolor alias_ok current_alias
    toplevel .alias -bg $bgcolor
    wm withdraw .alias
    frame .alias.f0 -bg $bgcolor
    label .alias.f0.l1 -bg $bgcolor -font $font -text {l̕ʖ}
    button .alias.f0.b1 -bg $bgcolor -font $font -text {ǉ} -command {
       set current_alias(alias) aliases
       set current_alias(window) .alias.f1.e
       new_alias
    }
    pack .alias.f0 -side top
    pack .alias.f0.l1 .alias.f0.b1 -side left

    frame .alias.f1 -bg $bgcolor
    pack .alias.f1 -side top -fill both -expand 1
    listbox .alias.f1.e -width 40 -height 5 -font $font -bg $bgcolor \
                          -yscrollcommand ".alias.f1.v set"
    scrollbar .alias.f1.v -orient vertical -bg $bgcolor -command ".alias.f1.e yview"
    grid .alias.f1.e -column 0 -row 0 -sticky nsew
    grid .alias.f1.v -column 1 -row 0 -sticky ns
    grid columnconfigure .alias.f1 0 -weight 100

    bind .alias.f1.e <Button-1> {
       set current_alias(line) [.alias.f1.e nearest %y]
       set current_alias(alias) aliases
       set current_alias(window) .alias.f1.e
       select_alias
    }

    frame .alias.f3 -bg $bgcolor
    label .alias.f3.l1 -bg $bgcolor -font $font -text {[OXg̕ʖ}
    button .alias.f3.b1 -bg $bgcolor -font $font -text {ǉ} -command {
       set current_alias(alias) mlaliases
       set current_alias(window) .alias.f2.e
       new_alias
    }
    pack .alias.f3 -side top
    pack .alias.f3.l1 .alias.f3.b1 -side left

    frame .alias.f2 -bg $bgcolor
    pack .alias.f2 -side top -fill both -expand 1
    listbox .alias.f2.e -width 40 -height 5 -font $font -bg $bgcolor \
                          -yscrollcommand ".alias.f2.v set"
    scrollbar .alias.f2.v -orient vertical -bg $bgcolor -command ".alias.f2.e yview"
    grid .alias.f2.e -column 0 -row 0 -sticky nsew
    grid .alias.f2.v -column 1 -row 0 -sticky ns
    grid columnconfigure .alias.f2 0 -weight 100
    bind .alias.f2.e <Button-1> {
       set current_alias(line) [.alias.f2.e nearest %y]
       set current_alias(alias) mlaliases
       set current_alias(window) .alias.f2.e
       select_alias
    }

    frame .alias.f4 -bg $bgcolor
    pack .alias.f4 -side top
    button .alias.f4.ok -bg $bgcolor -font $font -text {OK} -command {
                 set alias_ok 1
             }
    pack .alias.f4.ok -side left
}

set create_method(.al_popup) create_al_popup
proc create_al_popup {} {
    global font bgcolor
    toplevel .al_popup -bg $bgcolor
    wm withdraw .al_popup
    wm transient .al_popup
    frame .al_popup.f3 -bg $bgcolor
    pack .al_popup.f3 -side top
    label .al_popup.f3.l1 -bg $bgcolor -font $font -text {ʖ}
    label .al_popup.f3.l2 -bg $bgcolor -font $font -text {AhX}
    entry .al_popup.f3.e1 -bg $bgcolor -font $font -width 40
    entry .al_popup.f3.e2 -bg $bgcolor -font $font -width 40
    grid .al_popup.f3.l1 -column 0 -row 0 -sticky e
    grid .al_popup.f3.l2 -column 0 -row 1 -sticky e
    grid .al_popup.f3.e1 -column 1 -row 0 -sticky ew
    grid .al_popup.f3.e2 -column 1 -row 1 -sticky ew

    frame .al_popup.f4 -bg $bgcolor
    pack .al_popup.f4 -side top
    button .al_popup.f4.ok -bg $bgcolor -font $font -text {OK} -command {
                 apply_alias
                 popdown_panel .al_popup
             }
    button .al_popup.f4.cancel -bg $bgcolor -font $font -text {LZ} \
             -command {popdown_panel .al_popup}
    button .al_popup.f4.del -bg $bgcolor -font $font -text {폜} -command {
                 remove_alias
                 popdown_panel .al_popup
             }
    pack .al_popup.f4.ok .al_popup.f4.cancel .al_popup.f4.del -side left
}

proc edit_alias {} {
    global aliases mlaliases alias_ok
    popup_panel .alias
    .alias.f1.e delete 0 end
    .alias.f2.e delete 0 end
    foreach n [array name aliases] {
        .alias.f1.e insert end "$aliases($n)  $n"
    }
    foreach n [array name mlaliases] {
        .alias.f2.e insert end "$mlaliases($n)  $n"
    }
    tkwait variable alias_ok
    popdown_panel .alias
}

proc select_alias {} {
    global current_alias selected_alias_entry
    popup_panel .al_popup
    .al_popup.f3.e1 delete 0 end
    .al_popup.f3.e2 delete 0 end
    set al_line [$current_alias(window) get $current_alias(line)]
    .al_popup.f3.e1 insert end [lindex $al_line 0]
    .al_popup.f3.e2 insert end [lindex $al_line 1]
    set selected_alias_entry [lindex $al_line 1]
}

proc new_alias {} {
    global current_alias selected_alias_entry
    popup_panel .al_popup
    .al_popup.f3.e1 delete 0 end
    .al_popup.f3.e2 delete 0 end
    set selected_alias_entry {}
    set current_alias(line) end
}

proc apply_alias {} {
    global current_alias aliases mlaliases selected_alias_entry
    if [array_defined current_alias alias] {
        set nickname [.al_popup.f3.e1 get]
        set address [.al_popup.f3.e2 get]
        if {! [str_equal $current_alias(line) end]} {
            $current_alias(window) delete $current_alias(line) $current_alias(line)
        }
        $current_alias(window) insert $current_alias(line) "$nickname $address"
        eval [join [list "set " \
                         $current_alias(alias) \
                         "($address) $nickname"] {}]
        catch {
            eval [join [list "unset " \
                         $current_alias(alias) \
                         "($selected_alias_entry)"] {}]     
        }
    }
    update_ini
}

proc remove_alias {} {
    global current_alias aliases mlaliases selected_alias_entry
    if [array_defined current_alias alias] {
        $current_alias(window) delete $current_alias(line) $current_alias(line)
        eval [join [list "unset " \
                         $current_alias(alias) \
                         "($selected_alias_entry)"] {}]     
    }
    update_ini
}

#
# other procedures ----------------------------------------------------------
#

proc array_defined {arrayname elem} {
    upvar $arrayname a
    return [expr ! [catch {set a($elem)}]]
}

proc popup_geom {popup} {
    foreach var {x y w h px py pw ph} {
        set $var 0
    }
    regexp {([0-9]+)x([0-9]+)\+([0-9]+)\+([0-9]+)} [winfo geometry .] geom w h x y
    regexp {([0-9]+)x([0-9]+)\+([0-9]+)\+([0-9]+)} [winfo geometry $popup] geom pw ph px py
    if {$pw < 10} {
	set pw 300
	set ph 50
    }
    set x [expr $x+$w/2-$pw/2]
    set y [expr $y+$h/2-$ph/2]
    if {$x < 0} {set x 0}
    if {$y < 0} {set y 0}
    wm geometry $popup "+$x+$y"
}

proc popup_panel {panel} {
    global create_method
    if [catch {winfo id $panel}] {
	eval $create_method($panel)
    }
    wm deiconify $panel
    popup_geom $panel
    focus -force $panel
    if [winfo viewable $panel] {
        grab set $panel
    }
    update
}

proc popdown_panel {panel} {
    grab release $panel
    wm withdraw $panel
    focus -force .tl
    update
}

# dialog "Can't open file"
proc cantopen {file} {
    tk_messageBox -icon error -type ok -message "t@C $file J܂"
}

# dialog "You have no mail"
proc nomail {} {
    tk_messageBox -icon info -type ok -message {[͗Ă܂}
}

# mail save
proc m_save {file} {
    global current_mail mailbox
    if {$current_mail == -1} {
	tk_messageBox -icon error -type ok -message {܂[Ił}
        return 0
    } 
    if {[string length $file] == 0} {
	set file [tk_getSaveFile -initialdir $mailbox -filetypes {
                  {"Text file" {".txt"}} {"All file" {"*"}}}]
    }
    focus -force .tl
    if {[string length $file] > 0} {
	set catched [catch { set f [open $file a] }]
	if {$catched} {
	    cantopen $file
	    return 0
	}
	puts $f [.tx.text get 1.0 end]
	close $f
    }
    return 1
}


# Clear all mail
proc clear_mail {} {
    global n_mail
    .tb.e delete 0 end
    foreach w {num from to subject} { .tl.$w delete 0 end }
    .tx.text configure -state normal
    .tx.text delete 1.0 end
    .tx.text configure -state disabled
    set n_mail 0
    cleanup_image
}

# delete all images
proc cleanup_image {} {
    global image_list image_file_list
    foreach i $image_list {
        image delete $i
    }
    foreach i $image_file_list {
        f_unlink $i
    }
   set image_list {}
   set image_file_list {}
}

# Read mail file
proc read_mail {file} {
    global n_mail mails mailfile mail_status
    if {[string length $file] == 0} { return }
    if [catch {set f [open $file r]}] {
	nomail
	return
    }
    file stat $file f_info
    if {$f_info(size) > 0} {
        set mailfile $f
        popup_panel .progress
        progb_ratio 0
        progb_message {[ǂł܂}
        while {! [eof $f]} {
	    set pos [tell $f]
	    progb_ratio [expr "$pos.0" / $f_info(size)]
	    set mails($n_mail) [read_mailinfo $f 1]
	    set mail_status($n_mail) initial
	    insert_headline $n_mail $mails($n_mail)
	    incr n_mail
        }
        popdown_panel .progress
    } else { nomail }
}

#
# extract mail header of one mail
#
proc read_mailinfo {f through} {
    set in_header 1
    set header {}
    set mailpos [tell $f]
    set n_from 0
    foreach h {from to cc subject x-realname} {
        set minfo($h) {}
    }
    while {! [eof $f]} {
	gets $f line
	if { $in_header } {
	    if {[string length $line] == 0} {
		set in_header 0
		set prev_header $header
	    } elseif [regexp {^[ 	]+} $line] {
		append header $line
		continue
	    } else {
		set prev_header $header
		set header $line
	    }
	    if [head_match_nocase $prev_header "From:"] {
		set addr_name [extract_address [string range $prev_header 6 end]]
                set minfo(from) [lindex $addr_name 0]
                set minfo(x-realname) [lindex $addr_name 1]
	    } elseif [head_match_nocase $prev_header "To:"] {
		set minfo(to) [get_addrlist $prev_header]
	    } elseif [head_match_nocase $prev_header "Cc:"] {
		set minfo(cc) [get_addrlist $prev_header]
            } else {
                if [regexp {^([^:]+): (.*)$} $prev_header dummy field value] {
                    set minfo([string tolower $field]) $value
                }
            }
        }
        if {!$in_header} { break }
    }
    if {$through} {
        while {! [eof $f]} {
            set pos [tell $f]
            gets $f line
            if [head_match $line {From }] {
                seek $f $pos
                break
            }
        }
    }
    set minfo(pos) $mailpos
    return [array get minfo]
}

proc get_addrlist {field} {
    regexp {^[^:]+: (.*)$} $field dummy addrs
    set result {}
    foreach a [split $addrs ","] {
	regexp {[ ]*(.*)[ ]*} $a dummy a
	lappend result $a
    }
    return $result
}

#
# insert headline of mail
#
proc insert_headline {n mi} {
    incr n
    array set minfo $mi
    .tl.from insert $n [join $minfo(from)]
    .tl.to insert $n [join $minfo(to)]
    .tl.subject insert $n  [join $minfo(subject)]
    .tl.num insert $n $n
}

#
# show mail
#
set sepcount 0
proc show_mail {n} {
    global mailbox mails mailfile dir_separator mail_status sepcount sep_width
    global nimage image_list image_file_list tmpdir
    incr n -1
    # extract address part from To: and Cc: header, if needed
    array set thismail $mails($n)
    if [str_equal $mail_status($n) initial] {
        set mail_status($n) processed
        # Replace To:
        set addrs {}
        foreach addr $thismail(to) {
            lappend addrs [lindex [extract_address $addr] 0]
        }
        set thismail(to) $addrs
        # Replace Cc:
        set addrs {}
        foreach addr $thismail(cc) {
            lappend addrs [lindex [extract_address $addr] 0]
        }
        set thismail(cc) $addrs
        set mails($n) [array get thismail]
    }
    set nick [get_nickname $n]
    set savefile "$mailbox$dir_separator$nick.txt"
    .tb.e delete 0 end
    .tb.e insert 0 $savefile  
    seek $mailfile $thismail(pos)
    set nfrom 0
    .tx.text configure -state normal
    .tx.text delete 1.0 end
    set bodystart {}
    set boundary {}
    if {[array_defined thismail content-type] && \
        [head_match_nocase $thismail(content-type) "multipart"]} {
        regexp {boundary="([^"]+)"} $thismail(content-type) dummy boundary
        set sep_width [expr [winfo width .tx.text] - 10]
    }
    while {! [eof $mailfile]} {
	gets $mailfile line
	if [head_match $line "From "] {
	    if {$nfrom == 0} {
		incr nfrom
	    } else {
		break
	    }
	}
        if {[array_defined submail content-type]} {
          if {[head_match_nocase $submail(content-type) "image/gif"]} {
            set data $line
            while {1} {
                gets $mailfile line
                if [head_match $line "--$boundary"] { break }
                append data $line
            }
            set iname "tmpimg$nimage"
            set dname "$tmpdir${dir_separator}tmpimg$nimage.gif"
            decode_base64 $data $dname
            image create photo $iname -file $dname
            label .$iname -image $iname
            .tx.text window create end -window .$iname
            incr nimage
            lappend image_list $iname
            lappend image_file_list $dname
          }
        }
	if {[string length $bodystart] == 0 && [string length $line] == 0} {
	    set bodystart [.tx.text index end]
	}
        if {[string length $boundary] > 0 && [str_equal "--$boundary" $line]} {
            .tx.text window create end -create {
                 incr sepcount
                 frame .sep$sepcount -bg #ff0000 -height 2 -width $sep_width
            }
            .tx.text insert end "\n"
            array set submail [read_mailinfo $mailfile 0]
        } elseif {[string length $boundary] > 0 && [str_equal "--$boundary--" $line]} {
            # end of multipart message
            unset submail
        } else {
	    .tx.text insert end "$line\n"
        }
    }
    .tx.text configure -state disabled
    .tx.text yview $bodystart
}

#
# mail address ̋LqAAhXƖO؂蕪B
# ʂ { AhX O } ̌`̃XgƂĕԂB
#
proc extract_address {line} {
    if [regexp {(.*)<([^>]+)>(.*)} $line x p1 p2 p3] {
        set realname [join [list $p1 $p3]]
        set mailaddr $p2
    } else {
        set realname {}
        while {[regexp {([^()]*)\(([^)]*)\)(.*)} $line x p1 p2 p3]} {
            lappend realname $p2
            set line [join [list $p1 $p3] {}]
        }
        set mailaddr $line
    }
    regsub -all { +} $mailaddr {} mailaddr
    regsub -all { +} $realname { } realname
    return [list $mailaddr $realname]
}

#
# jbNl[̎擾DjbNl[gĊi[t@C肷D
# jbNl[̐ḰC
#   ETo: y Cc: ̓e mlalias Ɉv΁C̖O
#   EFrom: ̃AhX alias Ɉv΁C̖O
#   EFrom: ̃AhX̃[U
proc get_nickname {n} {
    global mails aliases mlaliases
    set a {}
    array set thismail $mails($n)
    set to_cc [concat $thismail(to) $thismail(cc)]
    foreach addr $to_cc {
	if {! [catch {set a $mlaliases($addr)}]} {
	    return $a
	}
    }
    set from $thismail(from)
    set addr [lindex [extract_address $from] 0]
    if [catch {set a $aliases($addr)}] {
	if {[string first @ $addr] >= 0} {
	    regsub {@.*} $addr {} nick
	    return $nick
	} else {
	    return $addr
	}
    }
    return $a
}

#
# bZ[W̏Ԃݒ肵CbZ[WԍɃ}[N
# bZ[W̏Ԃ mail_status( ) Ɋi[D
#   
proc article_status {n status} {
    global mail_status
    set mark {}
    switch $status {
	"saved" {set mark *}
	"deleted" {set mark x}
        "initial" {set mark { }}
        "processed" {set mark { }}
    }
    set ind [expr $n - 1]
    .tl.num delete $ind
    .tl.num insert $ind "$n$mark"
    set mail_status($ind) $status
    tl_sync
}
#
# send_mail: [𑗂
#    reply: ܑIĂ郁bZ[Wւ̕ԐM
#
proc send_mail {mode} {
    global current_mail mails wspoolfile username pophost genericfrom password
    global ezm_version editor uname_ok sendquery_ok fwdquery_ok debug
    set is_mode(reply) 0
    set is_mode(new) 0
    set is_mode(forward) 0
    set is_mode($mode) 1
    if {$is_mode(reply) || $is_mode(forward)} {
	if {$current_mail == -1} {
	    tk_messageBox -icon error -type ok -message {[Ił}
	    return
	}
        array set minfo $mails([expr $current_mail-1])
    }
    if [catch {set f [open $wspoolfile w]}] {
	cantopen $wspoolfile
	return
    }
    # wb_\
    if {$genericfrom} {
        regexp {[^.]\.(.*)} $pophost dummy domain
    } else {
        set domain $pophost
    }
    if {[string length $username] == 0 || [string length $password] == 0} { 
        query_userpass
        if {$uname_ok == 0} { return } 
    }
    if {$is_mode(reply)} {
        puts $f "From: $username@$domain"
        if [array_defined minfo reply-to] {
            puts $f [join [concat "To:" $minfo(reply-to)]]
	} else {
	    puts -nonewline $f "To: "
            puts $f [field_format $minfo(from)]
	    puts -nonewline $f "Cc: "
            puts $f [field_format [concat $minfo(to) $minfo(cc)]]
	}
        if [head_match_nocase $minfo(subject) {Re:}] {
            puts $f [join [concat "Subject:" $minfo(subject)]]
	} else {
            puts $f [join [concat "Subject: Re:" $minfo(subject)]]
	}
        puts $f "X-Mailer: EZM $ezm_version"
        puts $f "--text follows this line--"
    } elseif {$is_mode(new)} {
        puts $f "From: $username@$domain"
	popup_panel .sendquery
	focus -force .sendquery.p0
	tkwait variable sendquery_ok
        popdown_panel .sendquery
	if {$sendquery_ok} {
	    puts $f [join [list "To:" [.sendquery.p0 get]]]
	    puts $f [join [list "Cc:" [.sendquery.p1 get]]]
	    puts $f [join [list "Subject:" [.sendquery.p2 get]]]
	} else {
	    close $f
	    f_unlink $wspoolfile
	    return
	}
        puts $f "X-Mailer: EZM $ezm_version"
        puts $f "--text follows this line--"
    } elseif {$is_mode(forward)} {
        puts $f "Resent-From: $username@$domain"
        popup_panel .fwdquery
        focus -force .fwdquery.p0
        tkwait variable fwdquery_ok
        popdown_panel .fwdquery
        if {$fwdquery_ok} {
            puts $f [join [list "Resent-To:" [.fwdquery.p0 get]]]
	} else {
	    close $f
	    f_unlink $wspoolfile
	    return
	}
    }
    # p
    if {$is_mode(reply)} {
	set lines [.tx.text index end]
	set in_header 1
	for {set i 1} {$i <= $lines} {incr i} {
	    set line [.tx.text get $i.0 $i.end]
	    if {$in_header} {
		if {[string length $line] == 0} {
		    set in_header 0
                    if [array_defined minfo message-id] {
                        puts $f "In article $minfo(message-id),"
                    }
                    if [array_defined minfo x-realname] {
                        puts $f "\t$minfo(x-realname) said:"
                    } else {
                        puts $f "\t you said:"
                    }
		}
	    } else {
		puts $f "> $line"
	    }
	}
    } elseif {$is_mode(forward)} {
	set lines [.tx.text index end]
        set line [.tx.text get 1.0 1.end]
        if {! [head_match $line {From }]} {
            puts $f $line
        }
        for {set i 2} {$i <= $lines} {incr i} {
	    set line [.tx.text get $i.0 $i.end]
            puts $f $line
	}
    }
    close $f

    # GfB^̌Ăт
    set n_edit 1
    while {$n_edit > 0} {
	file stat $wspoolfile before_stat
	exec $editor $wspoolfile
	file stat $wspoolfile after_stat
        set ok yes
	if {$n_edit == 1 && $before_stat(mtime) == $after_stat(mtime)} {
            set ok [tk_messageBox -icon question -type yesno -default no\
                        -message {ҏWĂ܂łD̂܂ܑ܂H}]
        } else {
            set ok [tk_messageBox -icon question -type yesno -message {ĂłH}]
        }
        if [str_equal $ok yes] {
            # Ń[𑗂
            set send_ok [smtpsend $pophost $wspoolfile [expr ! $is_mode(forward)]]
            if [str_equal $send_ok ok] {
                set n_edit 0
            }
        } else { set send_ok dontsend }
        if [str_equal $send_ok dontsend] {
            set retry [tk_messageBox -icon question -type yesno -message {xҏW܂H}]
            if [str_equal $retry yes] {
                incr n_edit
                # xҏW
                continue
            } else {
                set n_edit 0
            }
        }
    }
    if {$debug} {.debug.win insert end "Out of while\n"}
    f_unlink $wspoolfile
}

# To:  Cc: Ȃǂ̃tB[h̓e܂Ԃ
proc field_format {addrs} {
    set len 4
    set result {}
    set naddr [llength $addrs]
    for {set i 0} {$i < $naddr} {incr i} {
	set a [lindex $addrs $i]
	append result $a
	incr len [string length $a]
	if {[expr $i+1] == $naddr} { 
	    set sep {}
	    set cr {}
	} else { 
	    set sep "," 
	    set cr "\n\t"
	}
	if {$len > 60} {
	    append result "$sep$cr"
	    set len 0
	} else {
	    append result $sep
	}
    }
    return $result
}

#
# I̍ۂɁCXv[XV
#
proc update_spool {} {
    global spoolfile tmpfile mails n_mail mail_status reading_spool
    if {! $reading_spool || ! [file exists $spoolfile]} {
        return
    }
    file stat $spoolfile f_stat
    if {$f_stat(size) == 0} { return }
    set reading_spool 0
    popup_panel .progress
    progb_message {[t@C`FbNĂ܂}
    progb_ratio 0

    # ꎞt@CI[v
    if [catch {set temp [open $tmpfile w]}] {
        popdown_panel .progress
        cantopen $tmpfile
        return
    }
    # spool I[v
    if [catch {set sp [open $spoolfile r]}] {
        popdown_panel .progress
        cantopen $spoolfile
        return
    }
    for {set i 0} {$i < $n_mail} {incr i} {
        set stat $mail_status($i)
        if {[str_equal $stat deleted] || [str_equal $stat saved]} {
            skip_one $sp {}
        } else {
            skip_one $sp $temp
        }
        progb_ratio [expr ("$i.0"+1)/$n_mail]
    }
    close $sp
    close $temp
    # ꎞt@CXv[ɏ㏑
    if [catch {set sp [open $spoolfile w]}] {
        popdown_panel .progress
        cantopen $spoolfile
        return
    }
    file stat $tmpfile f_stat
    if {$f_stat(size) > 0} {
        if [catch {set temp [open $tmpfile r]}] {
            cantopen $tmpfile
            popdown_panel .progress
            return
        }
        progb_message {[t@CXVĂ܂}
        progb_ratio 0.0
        set i 0
        while {! [eof $temp]} {
            gets $temp line
            puts $sp $line
            incr i
            if {$i == 200} {
                progb_ratio [expr [tell $temp].0/$f_stat(size)]
                set i 0
            }
        }
        close $temp
    }
    close $sp
    # ꎞt@C
    f_unlink $tmpfile
    popdown_panel .progress
}

proc skip_one {sp tmp} {
    set nfrom 0
    while {! [eof $sp]} {
        set pos [tell $sp]
        gets $sp line
        if [head_match $line {From }] {
            if {$nfrom} {
                seek $sp $pos
                break
            } else {
                incr nfrom
            }
        }
	if {[string length $tmp] > 0} { puts $tmp $line }
    }
}

# 
proc print_mail {} {
    global wspoolfile tcl_platform printcmd prog_base
    if [str_equal $tcl_platform(platform) unix] {
         m_save "|$printcmd"
    } else {
         set prn [select_printer]
         if {[string length $prn] > 0} {
             m_save $wspoolfile
             exec "${prog_base}ezmprint" $wspoolfile $prn
             f_unlink $wspoolfile
         }
    }
}

# l̏ݒt@C
proc update_ini {} {
    global rcfile pophost genericfrom username aliases mlaliases
    set rc_script {}
    set cmd(pophost) $pophost
    set cmd(genericfrom) $genericfrom
    set cmd(username) $username
    catch {
        set f [open $rcfile r]
        while {![eof $f]} {
            gets $f line
            if {[string length $line] == 0} {continue}
            foreach n [array name cmd] {
                if [regexp "set +$n +" $line] {
                    set line "set $n $cmd($n)"
                    set cmd($n) {}
                }
            }
            if [regexp {^[a-z]*alias .*} $line] { continue }
            append rc_script "$line\n"
        }
        close $f
    }
    if [catch {set f [open $rcfile w]}] {
        return
    }
    foreach n [array names cmd] {
        if {[string length $cmd($n)] > 0} {
           append rc_script "set $n $cmd($n)\n"
        }
    }
    foreach n [array name aliases] {
        append rc_script "alias $aliases($n) $n\n"
    }
    foreach n [array name mlaliases] {
        append rc_script "mlalias $mlaliases($n) $n\n"
    }
    puts -nonewline $f $rc_script
    close $f
}
 
############# NETWORK RELATED PROCEDURES #############################

#
# \PbgpoCfobO@\t
#
proc putsock {sock arg} {
    global debug
    puts $sock $arg
    flush $sock
    if {$debug} { 
	.debug.win insert end "<-$arg\n" 
	update
    }
}
proc getsock {sock s_var} {
    global debug
    upvar 1 $s_var var
    gets $sock var
    if {$debug} { 
	.debug.win insert end "->$var\n" 
	update
    }
}

#                            
# POP ֘A
#
proc pop_open {} {
    global username password popsock pophost
    if {[string length $password] == 0} {
        query_userpass
	if {[string length $username] == 0} {
	    return {}
	}
    }    
    if [catch { set sock [socket $pophost 110] }] {
	tk_messageBox -icon error -type ok -message [join [list $pophost {ւ̐ڑɎs܂}]]
	return {}
    }
    getsock $sock line
    if [head_match $line "-"] {
	tk_messageBox -icon error -type ok -message $line
	return {}
    }
    putsock $sock "USER $username"
    getsock $sock line
    if [head_match $line "-"] {
	tk_messageBox -icon error -type ok -message $line
	set username {}
	set password {}
	return {}
    }
    putsock $sock "PASS $password"
    getsock $sock line
    if [head_match $line "-"] {
	tk_messageBox -icon error -type ok -message $line
	set username {}
	set password {}
	return {}
    }
    set popsock $sock
}

proc pop_read_spool {} {
    global popsock spoolfile pophost
    popup_panel .progress
    progb_message [join [list $pophost {̓dq[`FbNĂ܂}]]
    progb_ratio 0
    putsock $popsock "STAT"
    getsock $popsock line
    set n [lindex [split $line] 1]
    set sp [open $spoolfile a]
    for {set i 1} {$i <= $n} {incr i} {
	putsock $popsock "RETR $i"
	#     next gets will get "+OK xxxx octets"
	getsock $popsock $line
	puts $sp "From pop@$pophost"
	set in_header 1
        set buffered 0
        set buffered_line {}
	while {1} {
            if {$buffered} {
                set line $buffered_line
                set buffered 0
            } else {
	        getsock $popsock line
            }
	    if [str_equal $line "."] { 
		break 
	    } elseif {[string length $line] == 0} {
		set in_header 0
	    }
	    if {$in_header} {
		if [regexp -nocase {=\?[A-Z0-9\-]+\?[BQ]\?} $line] {
		    # MIME decode
                    set hline $line
                    set cont 1
                    while {$cont} {
                        getsock $popsock line
                        if [regexp {^[	 ]} $line] {
                             append hline $line
                        } else { 
                             set buffered 1
                             set buffered_line $line
			     set cont 0
                        }
                    }
                    # remove space between encoded-words
                    regsub -all {\?=[	 ]+=\?} $hline "?==?" line
 		    set line [decode_mime $line]
                }
	    }
	    if [catch { puts $sp [jis2internal $line]}] {
                # disk full?
                tk_messageBox -type ok -icon error -message {Xv[t@C߂ȂȂ܂}
                while {![str_equal $line "."]} {
                    getsock $popsock line
                }
                putsock $popsock "QUIT"
                close $popsock
                close $sp
                set popsock {}
                popdown_panel .progress
                return
            }
	}
        putsock $popsock "DELE $i"
	#     next gets will get "+OK message xx was marked for deletion"
	getsock $popsock $line
        progb_ratio [expr "$i.0" / $n]
    }
    close $sp
    putsock $popsock "QUIT"
    close $popsock
    set popsock {}
    popdown_panel .progress
}

#
# SMTP֘A
#
proc smtp_check {code sock} {
    set line {}
    while {! [regexp {^[0-9]+ } $line]} {
	getsock $sock line
    }
    if {[lindex [split $line] 0] != $code} {
	tk_messageBox -icon error -type ok -message [join [list "SMTP error:" $line]]
	return 1
    }
    return 0
}

proc smtpsend {host file check_from} {
    global username tcl_platform
    # t@CJCwb_擾
    if [catch {set f [open $file r]}] {
	cantopen $file
	return error
    }
    array set minfo [read_mailinfo $f 0]
    if [str_equal $tcl_platform(platform) unix] {
        global internal
        set internal [file kanjiCode $file]
    }
    # From: 񂳂ĂȂ`FbN
    if {$check_from} {
        set smfrom [lindex [extract_address $minfo(from)] 0]
        if {! [mailaddr_equal $smfrom "$username@$host"]} {
            tk_messageBox -icon error -type ok -message \
		[join [list {Ȃ̃[AhX} "$username@$host" \
                {łD} $minfo(from) {ł͂܂D}]]
            return dontsend
        }
    }
    seek $f 0
    # SMTPڑ
    if [catch { set sock [socket $host 25] }] {
	tk_messageBox -icon error -type ok -message [join [list $host {ւ̐ڑɎs܂}]]
	return error
    }
    if [smtp_check 220 $sock] { 
	close $sock 
	return error
    }
    set myhost [gethostname]
    putsock $sock "HELO $myhost"
    if [smtp_check 250 $sock] {
	close $sock
	return error
    }
    # FROM ͎̃AhX
    putsock $sock "MAIL FROM: $username@$host"
    if [smtp_check 250 $sock] {
	close $sock
	return error
    }
    # 
    foreach rcpt [concat $minfo(to) $minfo(cc)] {
	putsock $sock "RCPT TO: $rcpt"
	if [smtp_check 250 $sock] {
	    close $sock
	    return error
	}
    }
    # eM
    putsock $sock "DATA"
    smtp_check 354 $sock
    update
    set in_header 1
    while {! [eof $f]} {
	gets $f line
	if [head_match $line {From }] {
	    set line [join [list ">" $line]]
	} elseif [str_equal "." $line] {
	    set line [join [list ">" $line]]
	} elseif {$in_header && [str_equal "--text follows this line--" $line]} {
	    set line {}
	    set in_header 0
	}
	if {$in_header} {
	    putsock $sock [encode_mime [internal2jis $line]]
	} else {
	    putsock $sock [internal2jis $line]
	}
    }
    putsock $sock "."
    smtp_check 250 $sock
    putsock $sock "QUIT"
    smtp_check 221 $sock
    close $sock
    return ok
}  
