#ifdef _WIN32
#include <stdio.h>
#include <windows.h>
#include <commdlg.h>
#include <string.h>
#include <sys/stat.h>

#undef DEBUG

typedef struct _textimg {
     char *str;
     unsigned short length;
     struct _textimg *next;
} textimg;

#define _NEW(type) ((type*)malloc(sizeof(type)))

#ifdef DEBUG
void printn(char *s, int n)
{
      int i;
      for (i = 0; i < n; i++)
           putchar(*(s++));
      puts("<cr>");
}
#endif

enum charstatus {cs_8bit=0, cs_16bit_1, cs_16bit_2};

/* check next character status (MS-KANJI only) */
enum charstatus nextstatus(enum charstatus current, char c)
{
     unsigned char uc = (unsigned char)c;
     switch (current) {
     case cs_8bit:
     case cs_16bit_2:
          if (uc < 0x80 || (0xa0 < uc && uc < 0xe0))
               return cs_8bit;
          else
               return cs_16bit_1;
     case cs_16bit_1:
          return cs_16bit_2;
     }
}

textimg *create_textimg(char *s,int maxchar,int *nlines)
{
     char *p = s;
     int n = 0;
     textimg *t = _NEW(textimg);
     textimg *text = t;
     enum charstatus cs = cs_8bit;

     t->str = p;
     if (nlines) *nlines = 0;

     while (*p) {
          cs = nextstatus(cs,*p);
          if (*p == '\n' || (n >= maxchar-2 && cs != cs_16bit_2)) {
              textimg *nt;
              t->length = (unsigned short)n;
#ifdef DEBUG
              printn(t->str,t->length);
#endif
              n = 0;
              if (*p == '\n')
                   p++;
              if (nlines) (*nlines)++;
              if (p == '\0')
                   t->next = 0;
              else {
                   nt = _NEW(textimg);
                   nt->str = p;
                   t->next = nt;
                   t = nt;
              }
          }
          else {
              p++;
              n++;
          }
     }
     return text;
}

void free_textimg(textimg *t)
{
     textimg *nt;
     while (t) {
          nt = t->next;
          free(t);
          t = nt;
     }
}

textimg *get_textimg(textimg *base, int line)
{
    while (line--) {
         base = base->next;
         if (base == 0) return 0;
    }
    return base;
}

#if 0
HDC GetPrinterDC()
{
     PRINTER_INFO_5 pinfo5[3];
     DWORD dwNeeded, dwReturned;

     if (EnumPrinters(/*PRINTER_ENUM_DEFAULT*/1,NULL,5,(LPBYTE)pinfo5,
         sizeof(pinfo5),&dwNeeded,&dwReturned))
          return CreateDC(NULL,pinfo5[0].pPrinterName,NULL,NULL);
     return 0;
}
#endif

BOOL PrintString (char *s,char *printer)
{
     static DOCINFO  di = { sizeof (DOCINFO), "mail -- ezmprint", NULL } ;
     HDC hDC ;
     /* static PAGESETUPDLG pgdlg; */
     BOOL            bSuccess ;
     int             xres,yres,xdpi,ydpi,xmargin, ymargin;
     int             yChar, iCharsPerLine, iLinesPerPage, iTotalLines, i,
                     iTotalPages, iPage, iLine, iLineNum ;
     TEXTMETRIC      tm ;
     textimg         *timg,*nimg;

#if 0 /* printer page setup dialog */
        memset(&pgdlg, 0, sizeof(PAGESETUPDLG));
        pgdlg.lStructSize = sizeof(PAGESETUPDLG);
        i = PageSetupDlg(&pgdlg);
        printf("PageSetupDialog returns %d\n",i);
#endif

     hDC = CreateDC(NULL,printer,NULL,NULL);
#ifdef DEBUG
     printf("printer DC = %d\n",hDC);
#endif
     if (hDC == 0)
         return FALSE;

     GetTextMetrics (hDC, &tm) ;
     yChar = tm.tmHeight + tm.tmExternalLeading ;

     xres = GetDeviceCaps (hDC, HORZRES);
     yres = GetDeviceCaps (hDC, VERTRES);
     xdpi = GetDeviceCaps (hDC, LOGPIXELSX);
     ydpi = GetDeviceCaps (hDC, LOGPIXELSY);
     xmargin = xdpi/2;  // 0.5 inch
     ymargin = ydpi/2;  // 0.5 inch
     yChar += ydpi/100; // adds 0.01inch space between lines
     iCharsPerLine = (xres-xmargin*2) / tm.tmAveCharWidth ;
     iLinesPerPage = (yres-ymargin*2) / yChar ;

     timg = create_textimg(s,iCharsPerLine,&iTotalLines);
     iTotalPages   = (iTotalLines + iLinesPerPage - 1) / iLinesPerPage ;
#ifdef DEBUG
     printf("%d chars/line\n",iCharsPerLine);
     printf("%d lines/page\n",iLinesPerPage);
     printf("%d total lines\n",iTotalLines);
     printf("%d total pages\n",iTotalPages);
#endif
     if (iTotalLines == 0)
          return TRUE ;

     if (StartDoc (hDC, &di) > 0) {
          for (iPage = 0 ; iPage < iTotalPages ; iPage++) {
               if (StartPage (hDC) < 0){
                    bSuccess = FALSE ;
                    break ;
               }

               for (iLine = 0 ; iLine < iLinesPerPage ; iLine++) {
                    iLineNum = iLinesPerPage * iPage + iLine ;

                    if (iLineNum > iTotalLines)
                         break ;
                    nimg = get_textimg(timg,iLineNum);
                    if (nimg)
                         TextOut (hDC, xmargin, ymargin+yChar * iLine,
                                  nimg->str,nimg->length);
               }

               if (EndPage (hDC) < 0) {
                    bSuccess = FALSE ;
                    break ;
               }
          }
     }
     else
          bSuccess = FALSE ;

     if (bSuccess)
          EndDoc (hDC) ;

     free_textimg(timg) ;
     DeleteDC (hDC) ;

     return bSuccess ;
}

char*
read_file_to_string(char *filename)
{
    struct stat st;
    char *buf,*p;
    FILE *f;

    if (stat(filename,&st) < 0) {
        return 0;
    }
    if ((buf = (char*)malloc(st.st_size+1)) == NULL) {
        return NULL;
    }
    f = fopen(filename,"r");
    fread(buf,st.st_size,sizeof(char),f);
    buf[st.st_size] = '\0';
    fclose(f);
    return buf;
}

main(int argc, char *argv[])
{
    char *printer,*str;
    if (argc < 3) exit(0);
    printer = argv[2];
    if ((str = read_file_to_string(argv[1])) == 0)
        exit(1);
    PrintString(str,printer);
    free(str);
    exit(0);
}
#endif
