#include <stdio.h>
#include <tcl.h>
#include <tk.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>

static char lbuf[4096];
Tcl_Interp *global_interp;
char *prog_base;

#ifdef _WIN32
#include "conv.h"

/* TCL command: jis2sj */
int 
jis2sj(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
        if (argc != 2) {
                Tcl_SetResult(interp,"usage: jis2sj string",TCL_STATIC);
                return TCL_ERROR;
        }
        conv(argv[1],lbuf,Code_A,Code_S);
        Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
        return TCL_OK;
}

/* TCL command: sj2jis */
int 
sj2jis(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
        if (argc != 2) {
                Tcl_SetResult(interp,"usage: sj2jis string",TCL_STATIC);
                return TCL_ERROR;
        }
        conv(argv[1],lbuf,Code_S,Code_N);
        Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
        return TCL_OK;
}

#endif

/* TCL command: gethostname */
int
tcl_gethostname(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    char hname[256];
    gethostname(hname,256);
    Tcl_SetResult(interp,hname,TCL_VOLATILE);
    return TCL_OK;
}

#ifndef _WIN32
/* TCL command: unlink */
int
tcl_unlink(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    /*extern int errno;*/
    if (argc != 2) {
	Tcl_SetResult(interp,"usage: unlink filename",TCL_STATIC);
	return TCL_ERROR;
    }
    if (unlink(argv[1]) == 0)
	return TCL_OK;
    switch (errno) {
    case EACCES:
	Tcl_SetResult(interp,"unlink: permission denied",TCL_STATIC);
	break;
    case EBUSY:
	Tcl_SetResult(interp,"unlink: file is busy",TCL_STATIC);
	break;
    case EFAULT:
	Tcl_SetResult(interp,"unlink: path too long",TCL_STATIC);
	break;
    case EINVAL:
	Tcl_SetResult(interp,"unlink: attempt to remove current directory",TCL_STATIC);
	break;
    case EIO:
	Tcl_SetResult(interp,"unlink: I/O error",TCL_STATIC);
	break;
    }
    return TCL_ERROR;
}

int
tcl_rename(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    /*extern int errno;*/
    if (argc != 3) {
	Tcl_SetResult(interp,"usage: rename filename1 filename2",TCL_STATIC);
	return TCL_ERROR;
    }
    if (rename(argv[1],argv[2]) == 0)
	return TCL_OK;
    switch (errno) {
    case EACCES:
	Tcl_SetResult(interp,"rename: permission denied",TCL_STATIC);
	break;
    case EBUSY:
	Tcl_SetResult(interp,"rename: file is busy",TCL_STATIC);
	break;
    case EFAULT:
	Tcl_SetResult(interp,"rename: path too long",TCL_STATIC);
	break;
    case EINVAL:
	Tcl_SetResult(interp,"rename: attempt to remove current directory",TCL_STATIC);
	break;
    case EIO:
	Tcl_SetResult(interp,"rename: I/O error",TCL_STATIC);
	break;
    default:
	Tcl_SetResult(interp,"rename: other error",TCL_STATIC);
	break;
    }
    return TCL_ERROR;
}

/* TCL command: mkdir */
int
tcl_mkdir(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
/*    extern int errno;  */
    if (argc != 2) {
	Tcl_SetResult(interp,"usage: mkdir filename",TCL_STATIC);
	return TCL_ERROR;
    }
    if (mkdir(argv[1],0700) == 0)
	return TCL_OK;
    switch (errno) {
    case EACCES:
	Tcl_SetResult(interp,"mkdir: permission denied",TCL_STATIC);
	break;
    case EBUSY:
	Tcl_SetResult(interp,"mkdir: file is busy",TCL_STATIC);
	break;
    case EFAULT:
	Tcl_SetResult(interp,"mkdir: path too long",TCL_STATIC);
	break;
    case EIO:
	Tcl_SetResult(interp,"mkdir: I/O error",TCL_STATIC);
	break;
    }
    return TCL_ERROR;
}
#endif /* _WIN32 */

/* TCL command: decode_mime */
int
tcl_decode_mime(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc != 2) {
	Tcl_SetResult(interp,"usage: decode_mime string",TCL_STATIC);
	return TCL_ERROR;
    }
    decodeMIME(argv[1],lbuf);
    Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
    return TCL_OK;
}

/* TCL command: encode_mime */
int
tcl_encode_mime(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc != 2) {
	Tcl_SetResult(interp,"usage: encode_mime string",TCL_STATIC);
	return TCL_ERROR;
    }
    encodeMIME(argv[1],lbuf);
    Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
    return TCL_OK;
}

/* TCL command: decode_base64 */
int
tcl_decode_base64(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc != 3) {
	Tcl_SetResult(interp,"usage: decode_base64 string file",TCL_STATIC);
	return TCL_ERROR;
    }
    if (decodeB2file(argv[1],argv[2]) < 0) {
        Tcl_SetResult(interp,"File error",TCL_STATIC);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/* TCL command: decode_quoted_printable */
int
tcl_decode_quoted_printable(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc != 3) {
	Tcl_SetResult(interp,"usage: decode_quoted_printable string file",TCL_STATIC);
	return TCL_ERROR;
    }
    if (decodeQ2file(argv[1],argv[2]) < 0) {
        Tcl_SetResult(interp,"File error",TCL_STATIC);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/* TCL command: head_match */
int
tcl_head_match(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    int len;
    if (argc < 3) {
	Tcl_SetResult(interp,"usage: head_match string string",TCL_STATIC);
	return TCL_ERROR;
    }
    len = strlen(argv[2]);
    sprintf(lbuf,"%d",!strncmp(argv[1],argv[2],len));
    Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
    return TCL_OK;
}

/* TCL command: head_match_nocase */
int
tcl_head_match_nocase(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    int len;
    if (argc < 3) {
	Tcl_SetResult(interp,"usage: head_match_nocase string string",TCL_STATIC);
	return TCL_ERROR;
    }
    len = strlen(argv[2]);
    sprintf(lbuf,"%d",!strncasecmp(argv[1],argv[2],len));
    Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
    return TCL_OK;
}

/* TCL command: str_equal */
int
tcl_str_equal(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc < 3) {
	Tcl_SetResult(interp,"usage: str_equal string string",TCL_STATIC);
	return TCL_ERROR;
    }
    sprintf(lbuf,"%d",!strcmp(argv[1],argv[2]));
    Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
    return TCL_OK;
}

/* TCL command: str_equal_nocase */
int
tcl_str_equal_nocase(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc < 3) {
	Tcl_SetResult(interp,"usage: str_equal_nocase string string",TCL_STATIC);
	return TCL_ERROR;
    }
    sprintf(lbuf,"%d",!strcasecmp(argv[1],argv[2]));
    Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
    return TCL_OK;
}

int
tcl_readsqeezedfile(Tcl_Interp *interp, char *filename)
{
    struct stat st;
    char *buf,*p;
    FILE *f;
    int res;

    if (stat(filename,&st) < 0) {
        Tcl_SetResult(interp,"Can't open sqeezed file",TCL_STATIC);
	return TCL_ERROR;
    }
    if ((buf = (char*)malloc(st.st_size)) == NULL) {
        Tcl_SetResult(interp,"no more memory",TCL_STATIC);
	return TCL_ERROR;
    }
    f = fopen(filename,"r");
    p = buf;
    while (fgets(lbuf,4096,f) != NULL) {
        int len = strlen(lbuf);
	char *q;
        sqeeze_line(lbuf);
	for (q = lbuf; *q; q++)
	    *(p++) = *q;
        *(p++) = '\n';
    }
    *p = '\0';
    res = Tcl_Eval(interp,buf);
    free(buf);
    return res;
}

#define strnequal(a,b,n) (strncasecmp(a,b,n)==0)
#define AddrChar(c)	(c == '@' || c == '.' || c == '!' || c == '%' || \
		     	 c == ':' || c == ',' || c == '-' || c == '_' || \
                         c == '/' || \
			 isalnum(c))


static int
mailaddrcmp(alias,addr)
char *alias,*addr;
{
	char *p,*q;
	int i,nl;

	p = strchr(alias,'@');
	if (p == NULL) {
		return strcmp(alias,addr);
	}
	nl = p-alias+1;
	if (!strnequal(alias,addr,nl))
		return -1;
	p++;
	q = strchr(addr,'@')+1;
	for (nl = 0; AddrChar(p[nl]); nl++);  
	for (;;) {
		if (strnequal(p,q,nl))
			return 0;
		if ((q = strchr(q,'.')) == NULL)
			return -1;
		q++;
	}
}

/* TCL command: mailaddr_equal */
int
tcl_mailaddr_equal(ClientData data, Tcl_Interp *interp, int argc, char *argv[])
{
    if (argc < 3) {
	Tcl_SetResult(interp,"usage: mailaddr_equal string string",TCL_STATIC);
	return TCL_ERROR;
    }
    sprintf(lbuf,"%d",!mailaddrcmp(argv[1],argv[2]));
    Tcl_SetResult(interp,lbuf,TCL_VOLATILE);
    return TCL_OK;
}


int
Tcl_AppInit(Tcl_Interp *interp)
{
	char evalbuf[256];

#ifdef _WIN32
        sprintf(evalbuf,"set tcl_library {%slib/tcl7.6}",prog_base);
        Tcl_Eval(interp,evalbuf);
        sprintf(evalbuf,"set tk_library {%slib/tk4.2}",prog_base);
        Tcl_Eval(interp,evalbuf);
#endif
	sprintf(evalbuf,"set prog_base {%s}",prog_base);
	Tcl_Eval(interp,evalbuf);

	global_interp = interp;

	if (Tcl_Init(interp) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if (Tk_Init(interp) == TCL_ERROR) {
	    return TCL_ERROR;
	}
	Tcl_StaticPackage(interp, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL);

	Tcl_CreateCommand(interp,"gethostname",tcl_gethostname,NULL,NULL);
#ifdef _WIN32
        Tcl_CreateCommand(interp,"jis2sj",jis2sj,NULL,NULL);
        Tcl_CreateCommand(interp,"sj2jis",sj2jis,NULL,NULL);
#else
	Tcl_CreateCommand(interp,"f_unlink",tcl_unlink,NULL,NULL);
	Tcl_CreateCommand(interp,"f_rename",tcl_rename,NULL,NULL);
	Tcl_CreateCommand(interp,"f_mkdir",tcl_mkdir,NULL,NULL);
#endif
	Tcl_CreateCommand(interp,"encode_mime",tcl_encode_mime,NULL,NULL);
	Tcl_CreateCommand(interp,"decode_mime",tcl_decode_mime,NULL,NULL);
	Tcl_CreateCommand(interp,"decode_base64",tcl_decode_base64,NULL,NULL);
	Tcl_CreateCommand(interp,"decode_quoted_printable",tcl_decode_quoted_printable,NULL,NULL);
	Tcl_CreateCommand(interp,"head_match",tcl_head_match,NULL,NULL);
	Tcl_CreateCommand(interp,"head_match_nocase",tcl_head_match_nocase,NULL,NULL);
	Tcl_CreateCommand(interp,"str_equal",tcl_str_equal,NULL,NULL);
	Tcl_CreateCommand(interp,"str_equal_nocase",tcl_str_equal_nocase,NULL,NULL);
	Tcl_CreateCommand(interp,"mailaddr_equal",tcl_mailaddr_equal,NULL,NULL);

        sprintf(evalbuf,"%sskel.sqz",prog_base);
        if (tcl_readsqeezedfile(interp,evalbuf) != TCL_OK)
		printf("%s\n",interp->result);
	return TCL_OK;
}

void
term_proc()
{
    Tcl_Eval(global_interp,"cleanup_image");
    Tcl_Eval(global_interp,"update_spool");
    exit(0);
}

main(int argc, char *argv[])
{
    int i;
    char *p;
    signal(SIGTERM,term_proc);
#ifdef SIGPIPE
    signal(SIGPIPE,term_proc);
#endif

#ifdef _WIN32
    p = argv[0];
    for (i = strlen(p)-1; i >= 0 && p[i] != '/'; --i);
    if (p[0] == '/' && p[1] == '/' && p[3] == '/') {
        /* drive-letter for GNUWIN32 */
        p++;
        p[0] = p[1];
        p[1] = ':';
        i--;
    }
    strncpy(lbuf,p,i+1);
    prog_base = strdup(lbuf);
#else /* UNIX */
    prog_base = PROG_BASE;
#endif

    Tk_Main(argc,argv,Tcl_AppInit);
    exit(0);
}

